/*
** Module   : ChuiWebSimulator.java
** Abstract : A WebSocket Chui Terminal simulator.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20131113 Created initial version.
** 002 MAG 20140114 Implements WebSockets protocol and communication.
** 003 MAG 20140204 Send text attributes via WebSockets. Add a Watchdog Timer 
**                  guardian and quit command.                   
** 004 MAG 20140211 Implements COLOR.
** 005 CA  20140206 Web socket timeout can't be read from the jetty threads (as these don't have a
**                  P2J context, so it must be passed at construction.
** 006 MAG 20140217 Refactoring web client redirection flaw.
** 007 MAG 20140219 Improve performances.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.util.*;

/**
 * Chui WebSocket terminal simulator.
 * Implements the WebSockets protocol to communicate with the JS terminal simulator.
 * The payloads are in 2 formats binary and JSON text.
 * <p>
 * Input messages:
 * 
 * Keyboard input.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x01  message id
 * 1       uu    keystroke int value
 * 2       ll    keystroke = uu << 8 | ll 
 * <p>
 * Output messages:
 * 
 * Clear screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x80  message id
 * <p>
 * Draw text on screen.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x81  message id
 * 1..n    data  data packets
 * 
 * structure of data packets:

 * offset  value content
 * ------------------------------------
 * 0       size  packet size
 * 1       row   row number
 * 2       col   column number
 * 3       attr  text attributes
 * 4       color FG/BG color pairs
 * 5..n    text  text characters, each character on 2 bytes (UTF-16)
 * <p>
 * Set cursor position.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x82  message id
 * 1       row    row
 * 2       col   column
 * <p>
 * Set cursor status.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x83  message id
 * 1       stat  status 1 = visible 
 *                      0 = hidden 
 * <p>
 * Play a beep message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x84  message id
 * <p>
 * Quit message.
 * 
 * offset  value content
 * ------------------------------------
 * 0       0x85  message id
 * <p>
 * Palette message is in JSON format as follow.
 * 
 *    {"c":0, "p":[color1, color2, ... color3]}
 * 
 * Example:
 *    {"c":0, "p":["#000000", "#FF0000", "#00FF00", "#0000FF", "#FFFFFF"]}
 * 
 */
@SuppressWarnings("serial")
@WebSocket(maxIdleTime = 900000, maxTextMessageSize = 4096)
public class ChuiWebSimulator
extends ChuiSimulator
{  
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ChuiWebSimulator.class.getName()); 
   
   /** keyboard stroke message */
   private static final byte MSG_KEY           = (byte) 0x01;

   /** clear screen message */
   private static final byte MSG_CLEAR         = (byte) 0x80;

   /** draw text on screen */
   private static final byte MSG_DRAW          = (byte) 0x81;
   
   /** set cursor position */
   private static final byte MSG_CURSOR_POS    = (byte) 0x82;
   
   /** set cursor status */
   private static final byte MSG_CURSOR_STATUS = (byte) 0x83;
   
   /** message beep */
   private static final byte MSG_BEEP          = (byte) 0x84;
   
   /** quit message */
   private static final byte MSG_QUIT          = (byte) 0x85;
   
   /** Color palette */
   private static final String COLOR_PALETTE = "{\"c\":0, \"p\":%s}"; 
   
   /** Keyboard input queue */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Keep cursor status */
   private boolean cursorVisible = true;
   
   /** Web socket timeout. */
   private final long timeout;
   
   /** Watchdog time out value */
   private final long watchdogTimeout;

   /** Watchdog instance */
   private WatchdogTimer watchdog;
   
   /** Color palette JSON message */
   private String paletteMessage;
   
   /** Number of rows */
   private int rows;
   
   /** Number of columns */
   private int columns;

   /** Screen contents. */
   private Cell[][] cache = null;
     
   /** Push messages worker thread */
   private PushMessagesWorker pushWorker;
   
   /**
    * Create a new chui web simulator.
    * 
    * @param    timeout
    *           Web socket timeout.
    * @param    watchdogTimeout
    *           Watchdog thread timeout.
    */
   public ChuiWebSimulator(long timeout, long watchdogTimeout)
   {
      // build palette message
      paletteMessage = buildPaletteMessage();
      this.timeout = timeout;
      this.watchdogTimeout = watchdogTimeout;
      
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      synchronized (keyBuffer)
      {
         // Stop watchdog timer
         stopWatchdog();
         // Get RemoteEndpoint reference.
         if (timeout > 0)
         {
            session.setIdleTimeout(timeout);
         }
         // Start push worker thread
         startPushWorker(session);
         // synchronize tty
         initRemoteTerminal();
      }
   }

   /**
    * This method is called when a text message is received.
    *
    * @param   message
    *          The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) 
   {
      // Not used.
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param   message
    *          Message content as an byte array.
    * @param   offset
    *          Message offset..
    * @param   length
    *          Message size.
    */
   @OnWebSocketMessage
   public void onMessage(byte[] message, int offset, int length)
   {
      if (length == 3 && message[offset] == ChuiWebSimulator.MSG_KEY)
      {
         int keystroke = (message[offset + 1] << 8) & 0xFF00;
         keystroke |= (message[offset + 2] & 0x00FF); 
         keyBuffer.offerLast(keystroke);         
      }
   }

   /**
    * This method is called when the connection is closed.
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      synchronized (keyBuffer)
      {
         // Stop push worker thread
         stopPushWorker();
         // Start watchdog timer
         startWatchdog(watchdogTimeout);
      }
      
      LOG.logp(Level.INFO,
               "ChuiWebSimulator.onClose",
               "",
               LogHelper.generate("Connection closed status=%d reason=%s", statusCode, reason));
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      session.close();
      
      LOG.logp(Level.SEVERE,
               "ChuiWebSimulator.onError",
               "",
               LogHelper.generate("WebSocket communication error!"),
               error);      
   }

   /**
    * Start watchdog timer.
    * 
    * @param   watchdogTimeout
    *          Watchdog timeout value.
    */
   private void startWatchdog(long watchdogTimeout)
   {
      if (watchdog == null)
      {
         watchdog = new WatchdogTimer(watchdogTimeout);
      }
   }

   /**
    * Stop watchdog timer.
    */
   private void stopWatchdog()
   {
      if (watchdog != null)
      {
         watchdog.kill();
         watchdog = null;
      }
   }

   /**
    * Start pushWorker thread.
    * 
    * @param   session
    *          An active link of communications with a Remote WebSocket end point.
    */
   private void startPushWorker(Session session)
   {
      if (pushWorker == null)
      {
         pushWorker = new PushMessagesWorker(session);
      }
   }

   /**
    * Stop pushWorker thread.
    */
   private void stopPushWorker()
   {
      if (pushWorker != null)
      {
         pushWorker.stopPushing();
         pushWorker = null;
      }
   }
   
   /**
    * Send a text message.
    * 
    * @param   message
    *          Text message.
    */   
   public void sendTextMessage(String message) 
   {
      synchronized (keyBuffer)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(message);
         }
      }
   }

   /**
    * Send a binary message.
    * 
    * @param   message
    *          Binary message.
    */   
   public void sendBinaryMessage(byte[] message) 
   {
      synchronized (keyBuffer)
      {
         if (pushWorker != null)
         {
            pushWorker.pushMessage(ByteBuffer.wrap(message));
         }
      }
   }

   /**
    * Whenever the main page is loaded on browser synchronize virtual terminal state.
    */
   private void initRemoteTerminal()
   {
      // clear cache
      cache = new Cell[rows][columns];
      colorPalette();
      // synchronize
      triggerRepaint(true);
      setCursorStatus(cursorVisible);
      if (cursorVisible)
      {
         setCursor(getCursorX(), getCursorY());
      }
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      return keyBuffer.isEmpty() ? -1 : keyBuffer.pollFirst();
   }
   
   /**
    * Initialize the instance using the given configuration.
    * Allocate screen cache. 
    *
    * @param   options
    *          The configuration for the terminal simulator.
    */   
   @Override
   public void configure(TerminalOptions options)
   {
      super.configure(options);
      
      synchronized (keyBuffer)
      {
         columns = options.columns;
         rows = options.rows;
      
         if (cache == null)
         {
            cache = new Cell[rows][columns];
         }
      }
   }

   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0. Send INIT message down to JS client.
    */   
   @Override
   public void clear()
   {
      super.clear();
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_CLEAR);
      sendBinaryMessage(message.toByteArray());
   }

   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_BEEP);
      sendBinaryMessage(message.toByteArray());         
   }
   
   /**
    *  Set the cursor visibility. Send SET_CURSOR_ON or SET_CURSOR_OFF message down to JS client.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequent repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visibility state.
    */   
   @Override
   public boolean setCursorStatus(boolean on) 
   throws IllegalStateException
   {
      cursorVisible = on;
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_CURSOR_STATUS);
      message.write(on ? 1 : 0);
      sendBinaryMessage(message.toByteArray());         
      // call super implementation
      return super.setCursorStatus(on);
   }

   /**
    * Position the cursor. Send SET_CURSOR_POS message down to JS client.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean setCursor(int x, int y)
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_CURSOR_POS);
      message.write(y);
      message.write(x);
      // send message
      sendBinaryMessage(message.toByteArray());         
      // call super implementation
      return super.setCursor(x, y);
   }
   
   /**
    * Put cell attributes and character into the message buffer.
    * 
    * @param   message
    *          Message buffer.
    * @param   cell
    *          Cell object.
    */
   private void putCellAttributes(ByteArrayOutputStream message, Cell cell)
   {
      // attributes
      Color cellColor = cell.color();
      if (cellColor == null)
      {
         // attributes
         message.write(0);
         // color
         message.write(0);                        
      }
      else
      {
         // attributes
         message.write(cellColor.attribute());
         // color
         int color = cellColor.color();
         if (color == 0)
         {
            message.write(0);               
         }
         else
         {
            int bg = color & 0x07;
            int fg = (color >>> 12) & 0x70;
            message.write(fg | bg);
         }               
      }
      // character
      putCellCharacter(message, cell);
   }

   /**
    * Put cell character into the message buffer.
    * 
    * @param   message
    *          Message buffer.
    * @param   cell
    *          Cell object.
    */
   private void putCellCharacter(ByteArrayOutputStream message, Cell cell)
   {
      // character
      char c = cell.character();
      message.write((c & 0xFF00) >> 8);
      message.write(c & 0x00FF);
   }
   
   /**
    * Should be implemented by sending the screen status to the remote end point as a refresh
    * screen operation. The command is part of the communication protocol.
    * The colors and attributes messages are cached and performances are significantly improved.   
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      synchronized (keyBuffer)
      {
         // message
         ByteArrayOutputStream message = new ByteArrayOutputStream();
         message.write(ChuiWebSimulator.MSG_DRAW);
         
         // packet
         ByteArrayOutputStream packet = new ByteArrayOutputStream();

         // get screen data
         Cell[][] data = getScreenData();
               
         for (int row = 0; row < rows; row++)
         { 
            // attributes of previous cell
            Color previous = null;
            
            for (int col = 0; col < columns; col++)
            {
               Cell cell = data[row][col];
               // check if cell has changes
               if (cell != null && !cell.equals(cache[row][col]))
               {
                  cache[row][col] = cell;
                  Color color = cell.color(); 
                  // check for attributes changes
                  if (color != null && color.equals(previous))
                  {
                     putCellCharacter(packet, cell);
                  }
                  else
                  {
                     previous = color;
                     // add packet to the message
                     if (packet.size() > 0)
                     {
                        byte[] payload = packet.toByteArray();
                        payload[0] = (byte) payload.length;
                        message.write(payload, 0, payload.length);
                     }
                     // new message
                     packet.reset();    // reuse buffer
                     packet.write(0);   // size
                     packet.write(row); // line number                     
                     packet.write(col); 
                     putCellAttributes(packet, cell);
                  }
               }
               else
               {
                  previous = null;
               }
            }
            
            // add last packets if exists
            if (packet.size() > 0)
            {
               byte[] payload = packet.toByteArray();
               payload[0] = (byte) payload.length;
               message.write(payload, 0, payload.length);               
            }
            // send the message if exists
            if (message.size() > 1)
            {
               sendBinaryMessage(message.toByteArray());
            }
         }
      }
   }
   
   /**
    * Remote application is about to end. Notify the JS client about that.
    * The client should be redirected to referrer address.
    */
   public void quit()
   {
      ByteArrayOutputStream message = new ByteArrayOutputStream();
      message.write(ChuiWebSimulator.MSG_QUIT);
      sendBinaryMessage(message.toByteArray());         
   }
         
   /**
    * Build color palette message. This message is the same
    * over the user session so it's build and store on constructor.
    * The message is send whenever a WebSocket connection is open.
    *
    * @return  A string representing the JSON format of color palette.
    */
   private String buildPaletteMessage()
   {
      // build palette
      StringBuilder palette = new StringBuilder("[");      
      for (ColorPalette colorPalette : ColorPalette.values())
      {
         if (palette.length() != 1)
         {
            palette.append(",");
         }
         palette.append("\"").append(colorPalette.asWebColor()).append("\"");
      }      
      palette.append("]");      
      // JSON format
      return String.format(COLOR_PALETTE, palette.toString());
   }
   
   /**
    * Send color palette down to JS client.
    * This is done whenever a WebSocket connection is open.
    */
   private void colorPalette()
   {      
      // send palette
      sendTextMessage(paletteMessage);
   }
}